package com.jz.dm.service.request;

import com.alibaba.csp.sentinel.annotation.SentinelResource;
import com.alibaba.fastjson.JSONObject;
import com.jz.common.utils.HttpsUtils;
import com.jz.common.utils.RedisUtils;
import com.jz.dm.common.constant.LoggingConstants;
import com.jz.dm.common.enums.GatewayResultCode;
import com.jz.dm.common.enums.apiInterface.ApiInfoOutTypeEnum;
import com.jz.dm.common.enums.auth.AuthModeEnum;
import com.jz.dm.common.exception.GatewayException;
import com.jz.dm.common.util.OpenApiRequest;
import com.jz.dm.common.util.OpenApiResponse;
import com.jz.dm.common.util.stream.HttpDownload;
import com.jz.dm.gateway.OpenApiService;
import com.jz.dm.models.domian.ApiAuth;
import com.jz.dm.models.domian.ApiInterface;
import com.jz.dm.service.ApiInterfaceService;
import com.jz.dm.service.AuthService;
import com.jz.dm.web.annotation.AccessLimit;
import com.jz.dm.web.annotation.ApiLogAspect;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import java.util.Map;

/**
 * @author ZC
 * @PACKAGE_NAME: com.jz.dm.service.request
 * @PROJECT_NAME: jz-dm-parent
 * @NAME: ApiQueryService
 * @DATE: 2020-12-23/18:10
 * @DAY_NAME_SHORT: 周三
 * @Description:
 **/
@Service("apiQueryService")
@Slf4j
public class ApiQueryService extends ApiParamVerify implements OpenApiService {

    @Override
    public String getOpenApiMethod() {
        return "query";
    }

    @Override
    public String getOpenApiVersion() {
        return "v1.0.0";
    }

    @Autowired
    private ApiInterfaceService apiInterfaceService;
    @Autowired
    private AuthService authService;
    @Autowired
    private RedisUtils redisUtils;
    @Autowired
    private HttpsUtils httpsUtils;

    @Value("${data.bank.balanceUrl}")
    private String balanceUrl;

    /**
     * API请求逻辑处理
     *
     * @param request
     * @param response
     */
    @Override
    @ApiLogAspect(description = "API请求日志")
    @AccessLimit(limit = 10000, sec = 1)
    @SentinelResource(value = "api.gateway", fallback = "fallbackGateway")
    @Transactional(rollbackFor = Exception.class, propagation = Propagation.REQUIRES_NEW)
    public void doService(OpenApiRequest request, OpenApiResponse response) {
        JSONObject parameter = JSONObject.parseObject(request.getOpenApiParams());
        String authCode = parameter.getString("authCode");
        if (StringUtils.isBlank(authCode)) {
            throw new GatewayException(GatewayResultCode.ILLEGAL_REQUEST);
        }
        verifyApiInterface(apiInterfaceService.getApiInfo(request.getApiKey()));
        ApiAuth apiAuth = authService.getAuthInfo(authCode);
        verifyAuth(apiAuth);

        String reqParams = parameter.getString("reqParams");
        Map paramMap = null;
        if (StringUtils.isNotBlank(reqParams)) {
            paramMap = (Map) JSONObject.parseObject(reqParams);
        }
        //取出缓存数据
        String redisReqParam = redisUtils.get(request.getApiKey());
        boolean bResult = false;
        try {
            if (StringUtils.isNotBlank(redisReqParam)) {//redis中存在
                //解析出API制作成功时的参数配置
                JSONObject jsonObject = JSONObject.parseObject(redisReqParam);
                String targetUrl = jsonObject.getString("targetUrl");
                String outputType = jsonObject.getString("outputType");
                bResult = rangRequestTarget(outputType, targetUrl, paramMap, response);
            } else {//不存在查询数据库
                ApiInterface apiInterface =
                        apiInterfaceService.getReqTargetInfo(request.getApiKey());
                if (null == apiInterface) {
                    throw new GatewayException(GatewayResultCode.REQUEST_INFO_UNEXIST);
                }
                bResult = rangRequestTarget(apiInterface.getOutputType(),
                        apiInterface.getTargetUrl(), paramMap, response);
            }
            //调用成功请求数据银行扣款
            if (AuthModeEnum.POWER_CALL_MODE.name().equals(apiAuth.getAuthMode())) {
                notifierMinusMoney(parameter, bResult);
                //按次调用时处理(处理为已调用)
                authService.updateApiAuthStatus(apiAuth);
            }
        } catch (Exception ex) {
            if (ex instanceof GatewayException) {
                //调用失败回调数据银行解冻
                if (AuthModeEnum.POWER_CALL_MODE.name().equals(apiAuth.getAuthMode())) {
                    notifierMinusMoney(parameter, bResult);
                }
                throw (GatewayException) ex;
            }
            log.error("~~~~~~~~~~~~~~~请求api信息异常~~~~~~~~~~~~~");
            log.error("异常信息:{}", ex.getMessage());
            response.setCode(GatewayResultCode.ILLEGAL_REQUEST.getCode());
            response.setMsg(GatewayResultCode.ILLEGAL_REQUEST.getMsg());
        }
    }


    /**
     * 远程请求获取数据
     *
     * @param outputType
     * @param targetUrl
     * @param param
     * @param response
     */
    private boolean rangRequestTarget(String outputType, String targetUrl,
                                      Map<String, String> param, OpenApiResponse response) {
        if (StringUtils.isBlank(outputType)) {
            outputType = ApiInfoOutTypeEnum.JSON.name();
        }
        if (StringUtils.isBlank(targetUrl)) {
            throw new GatewayException(GatewayResultCode.ILLEGAL_REQUEST);
        }
        if (ApiInfoOutTypeEnum.FLOW.name().equals(outputType)) {//文件流形式请求
            StringBuilder builder = new StringBuilder();
            builder.append(targetUrl).append("?");
            for (Map.Entry<String, String> entry : param.entrySet()) {
                String key = entry.getKey();
                String value = entry.getValue();
                builder.append(key).append(LoggingConstants.AND_EQUAL)
                        .append(value).append(LoggingConstants.AND_SPILT);
            }
            String baseUrl = builder.substring(0, builder.length() - 1);
            HttpDownload.download(baseUrl);
            response.setCode(GatewayResultCode.SUCCESS.getCode());
            response.setMsg(GatewayResultCode.SUCCESS.getMsg());
            return true;
        } else if (ApiInfoOutTypeEnum.JSON.name().equals(outputType)) { //json格式请求
            response.setAttribute("调用成功!");
            response.setMsg(GatewayResultCode.SUCCESS.getMsg());
            response.setCode(GatewayResultCode.SUCCESS.getCode());
            /*String respResult = httpsUtils.doGet(targetUrl, param);
            JSONObject resp = JSONObject.parseObject(respResult);
            if (null != resp && 200 == resp.getInteger("code")) {
                response.setCode(GatewayResultCode.SUCCESS.getCode());
                response.setMsg(GatewayResultCode.SUCCESS.getMsg());
                response.setAttribute(resp.getString("data"));
                return true;
            } else {
                log.error("~~~~~~~~~~~~~~~~远程请求异常~~~~~~~~~~~~~~~~~");
                throw new GatewayException(GatewayResultCode.DISTANCE_REQUEST_EXCEPTION);
            }*/
        } else {
            throw new GatewayException(GatewayResultCode.OUTPUT_TYPE_EXCEPTION);
        }
        return false;
    }


    /**
     * 通知扣款
     *
     * @param parameter
     * @param bResult
     */
    private void notifierMinusMoney(JSONObject parameter, boolean bResult) {
        Integer assetsId = parameter.getInteger("assetsId");
        Integer userId = parameter.getInteger("userId");
        String dataPrice = parameter.getString("dataPrice");
        JSONObject jsonReq = new JSONObject();
        jsonReq.put("assetsId", assetsId);
        jsonReq.put("userId", userId);
        jsonReq.put("dataPrice", dataPrice);
        jsonReq.put("callStatus", bResult);//true 调用成功 扣款 false 调用失败,解冻金额
        String responseResult = httpsUtils.submitPost(balanceUrl+"/mall/financeCustomerAssets/unfreezeMoney", jsonReq.toString());
        JSONObject paramsResult = JSONObject.parseObject(responseResult);
        if (null != paramsResult) {
            if (200 != paramsResult.getInteger("code")) {
                log.info("~~~~~~~~~~~~~~~调用数据银行扣款失败~~~~~~~~~~~~~");
                throw new GatewayException(GatewayResultCode.CALL_AMOUNT_NOT_ENOUGH);
            }
        }
    }

    /**
     * 限流返回方法
     *
     * @param request
     * @param response
     */
    public void fallbackGateway(OpenApiRequest request, OpenApiResponse response) {
        log.info("用户请求过于频繁触发限流接口:ApiKey为:{}" + request.getApiKey());
        throw new GatewayException(GatewayResultCode.RATE_LIMIT_EXCEEDED);
    }
}
